/* A task, according to this task server, consists of an address
 * space, and one or more threads, and a pager.
 */

#include <l4.h>

#include "task.h"

/* For simplicity, use fix sizes, reallocation is easy to add later. */
#define TASK_MAX 1000
#define SUBSCRIPTIONS_MAX 10000

/* Place these at some random high address */
#define KIP_LOCATION (0xC000000UL)
#define UTC_LOCATION (0xD000000UL)

struct task_object;
struct subscription_object;
struct thread_object;

/* Thread that is allowed to manage a task's accounting id. */
static l4_thread_id_t master;
static l4_fpage_t kip_fpage;

struct task_object
{
  enum { TASK_DEAD, TASK_ZOMBIE, TASK_ALIVE } state;
  
  /* For now, only a single accounting id */
  l4_word_t accounting_id;

  /* First subscription for this task */
  unsigned subscription_list;
  unsigned thread_list;
};

/* A task id is an index into this array. Entry 0 is not used. */
static struct task_object task[TASK_MAX];

static void
task_init(void)
{
  unsigned i;
  for (i = 0; i<TASK_MAX; i++)
    task[i].state = TASK_DEAD;

  master = l4_nilthread;
  /* The kip is always one page, right? */
  kip_fpage = l4_fpage(KIP_LOCATION,
		       l4_page_size_mask());
}

static unsigned
task_alloc_id(void)
{
  /* Inefficient linear search */
  unsigned i;
  for (i = 1; i< TASK_MAX; i++)
    if (task[i].state == TASK_DEAD)
      return i;

  return 0;
}

/* A subscription id is an index into this array. Entry 0 is not
 * used. */
struct subscription_object
{
  unsigned task_id;
  unsigned next; /* Next subscription for the same object */

  /* task id of subscriber */
  unsigned subscriber;
  
  /* Thread to send notifications to. l4_nilthread means that the
   * subscriber just wants to keep a reference to the thread, without
   * getting any subscriptions */
  l4_thread_id_t target;
};

static struct subscription_object subscription[SUBSCRIPTIONS_MAX];

/* Operations. For now, we have a single thread and no locks. */
static unsigned
task_create(l4_thread_id_t caller, l4_thread_id_t pager, unsigned max_threads)
{
  unsigned task_id;
  unsigned caller_task;
  l4_thread_id_t thread_id;
  l4_word_t old;

#if 0 
  unsigned parent_subscription;
  unsigned child_subscribtion;
#endif
  
  if (l4_is_local_id(caller))
    return 0;
  
  caller_task = l4_version(caller);
  if (!caller_task || caller_task > TASK_MAX)
    return 0;
  
  task_id = task_alloc_id();
  if (!task_id)
    return 0;

  thread_id = l4_global_id(1, task_id);

  if (!l4_thread_control(thread_id, thread_id, thread_id,
			 l4_nilthread, (void *) UTC_LOCATION))
    return 0;

  if (!l4_space_control(thread_id, 0,
			kip_fpage,
			l4_fpage(UTC_LOCATION,
				 /* FIXME: Alignment, and sanity
				    checking */
				 max_threads * l4_utcb_area_size()),
			l4_anythread, &old))
    return 0;

  /* Ok, we have create a thread and an address space. Let the
   * supplied pager do the rest. */

  if (l4_thread_control(thread_id, thread_id, l4_nilthread,
			pager, (l4_word_t *) -1))
    return 0;

  /* Book keeping. */
  task[task_id].state = TASK_ALIVE;
  task[task_id].accounting_id = task[caller_task].accounting_id;

  /* FIXME: Old and new task should get references to each other */
  task[task_id].subscription_list = 0;

  /* FIXME: Book keeping of the threads is not yet implemented. */
  task[task_id].thread_list = 0;
    
  return task_id;
}

/* In the reply, use label 1 for success, and 0 for failure */
int
main(void)
{
  task_init();
  
  for (;;)
    {
      l4_msg_tag_t tag;
      l4_thread_id_t caller;

      tag = l4_wait(&caller);
      if (l4_ipc_failed(tag))
	continue;
      
      switch(l4_label(tag))
	{
	case TASK_CREATE:
	  {
	    l4_thread_id_t pager;
	    l4_word_t max_threads;
	    unsigned task_id;

	    if (l4_untyped_words(tag) != 2
		|| l4_typed_words(tag))
	      goto fail;
	    
	    l4_store_mr(1, &pager.raw);
	    l4_store_mr(2, &max_threads);
	    task_id = task_create(caller, pager, max_threads);

	    if (!task_id)
	      goto fail;

	    tag = l4_msg_tag();
	    tag.label = 1;
	    tag.typed = 1;

	    l4_set_msg_tag(tag);
	    l4_load_mr(1, task_id);

	    break;
	  }
	case TASK_MASTER:
	  if (l4_untyped_words(tag) != 2
	      || l4_typed_words(tag))
	    goto fail;
	  if (l4_is_thread_equal(master, l4_nilthread))
	    /* There can be only one */
	    goto fail;

	  master = caller;
	  tag = l4_msg_tag();
	  tag.label = 1;
	  l4_set_msg_tag(tag);

	  break;
	  
	default:
	fail:
	  tag = l4_msg_tag();
	  tag.label = 0;
	}
      /* Ignore value */
      (void) l4_send_timeout(caller, l4_zero_time);
    }
}
